/******************************************************************************

    Copyright © 2008-2013 Brendt Wohlberg

    This program is free software; you can redistribute it and/or modify
    it under the terms of version 2 of the GNU General Public License at
    http://www.gnu.org/licenses/gpl-2.0.txt.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    Most recent modification: 30 December 2013

******************************************************************************/

#include <math.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include "gpsfmt.h"


/*****************************************************************************
 Convert radians to degrees and seconds.
 *****************************************************************************/
double radtodegsec(double rad) {
  const double dgrd = 180.0/M_PI;
  double deg, d, s;
  
  deg = dgrd*rad;
  d = (deg>=0?1.0:-1.0)*floor(fabs(deg));
  s = (deg - d) * 60.0;
  return 100.0*d + s;
}


/*****************************************************************************
 Round floating point value to 1 decimal place.
 *****************************************************************************/
float round1p(float x) {
  return roundf(10.0f*x)/10.0f;
}


/*****************************************************************************
 Print header for nmea log file.
 *****************************************************************************/
void print_hdr_nmea(FILE *stream, const char* btas) {
  char hstr[256] = {0};
  char bs[13] = {0};
#ifdef EMULATE_RTK_TOOL
  char *emflg = "-rtkfmt";
#else
  char *emflg = "";
#endif

  if (btas != NULL) {
    char bc[6][3] = {{0}};
    
    if (sscanf(btas, "%2s:%2s:%2s:%2s:%2s:%2s", bc[0], bc[1], bc[2], 
	       bc[3], bc[4], bc[5]) == 6)
      sprintf(bs, "%2s%2s%2s%2s%2s%2s", bc[0], bc[1], bc[2], bc[3], 
	      bc[4], bc[5]);
  }

  if (bs[0] == 0)
    sprintf(hstr, "$PRTK,RTKGPS,%s%s*", PACKAGE_VERSION, emflg);
  else 
    sprintf(hstr, "$PRTK,RTKGPS,%s%s,%12s*", PACKAGE_VERSION, emflg, bs);

  fprintf(stream, "%s%02X\r\n", hstr, string_checksum(hstr));
}


/*****************************************************************************
 Print log data fxp for file lgfl to the indicated output stream.
 *****************************************************************************/
void print_log_nmea(FILE *stream, const logfile_t *lfp, const gps_fix_t *fxp,
		    const float *gcp) {
  int n, i, j;
  char *pgga, *pgsa, *pgsv, *prmc, *prtd;
  char gga[256], gsa[256], gsv[256], rmc[256], rtd[256];
  char time[32];
  char alt[32];
  char vel[16];
  char sts[16];
  char crs[8];
  char ltd, lnd;
  char fxql[2] = "3";
  double lat, lon;
  double pdop = 0.0, hdop = 0.0, vdop = 0.0; 

  /* Loop over all fixes */
  for (n = 0; n < lfp->nfix; n++) {

    /* Construct formatted time string */
#ifdef EMULATE_RTK_TOOL
    sprintf(time, "%02d%02d%02d.000", fxp[n].hour, fxp[n].min, fxp[n].sec);
#else
    sprintf(time, "%02d%02d%02d.00", fxp[n].hour, fxp[n].min, fxp[n].sec);
#endif

    /* Compute latitude and longitude (and North/South/East/West
       indicators) in required formats */
    lat = fabs(radtodegsec(fxp[n].lat));
    ltd = (fxp[n].lat >= 0)?'N':'S';
    lon = fabs(radtodegsec(fxp[n].lng));
    lnd = (fxp[n].lng < 0)?'W':'E';

    /* Construct altitude description string depending on available
       information */
    if (lfp->fxtyp > 0) {
#ifdef EMULATE_RTK_TOOL
      (void)gcp;
      sprintf(alt, "%.1f,M,0.0,M", round1p(fxp[n].alt));
#else
      if (gcp != NULL)
	sprintf(alt, "%.1f,M,%.1f,M", round1p(fxp[n].alt-gcp[n]),
		round1p(gcp[n]));
      else
        sprintf(alt, "%.1f,M,,", round1p(fxp[n].alt));
#endif
    } else {
      sprintf(alt, ",,,");
    }

    /* Construct velocity description string depending on available
       information */
    if (lfp->fxtyp > 1)
      sprintf(vel, "%06.2f", 0.539956803f*fxp[n].vel);
    else
      vel[0] = '\0';

    /* Construct substrings for number of satellites/hdop and course
       depending on available information */
    if (lfp->fxtyp > 3) {
      short nsat;

      /* Fix quality indicator for GPGSA sentence. '2' and '3'
	 indicate 2D and 3D fixes respectively. '1' indicates no fix,
	 but is irrelevant here since logs are not written in this
	 case. */
      fxql[0] = (fxp[n].fstt & 0x10)?'3':'2';
      /* Number of satellites tracked field from GPGGA sentence. */
      nsat = fxp[n].nsat >> 4;
      /* Position dilution of precision */
      pdop = fxp[n].pdop / 100.0;
      /* Horizontal dilution of precision */
      hdop = fxp[n].hdop / 100.0;
      /* Vertical dilution of precision */
      vdop = fxp[n].vdop / 100.0;
      sprintf(sts, "%d,%.1f", nsat, hdop);
      sprintf(crs, "%.2f", fxp[n].crse);
    } else {
      fxql[0] = '\0';
      sprintf(sts, ",");
      crs[0] = '\0';
    }

    /* NMEA sentence prefixes; if a bad fix is detected, the standard
       NMEA prefix follows $PRTK,BADFIX so that applications that
       parse the NMEA data do not make use of those fixes. */
    pgga = (fxp[n].unkw0 == 0)?"$GPGGA":"$PRTK,BADFIX,GPGGA";
    pgsa = (fxp[n].unkw0 == 0)?"$GPGSA":"$PRTK,BADFIX,GPGSA";
    pgsv = (fxp[n].unkw0 == 0)?"$GPGSV":"$PRTK,BADFIX,GPGSV";
    prmc = (fxp[n].unkw0 == 0)?"$GPRMC":"$PRTK,BADFIX,GPRMC";
    prtd = (fxp[n].unkw0 == 0)?"$RTDIST":"$PRTK,BADFIX,RTDIST";

    /* Write GPGGA sentence */
#ifdef EMULATE_RTK_TOOL
    sprintf(gga, "%s,%s,%09.4f,%c,%010.4f,%c,1,%s,%s,,0000*", 
	    pgga, time, lat, ltd, lon, lnd, sts, alt);
#else
    sprintf(gga, "%s,%s,%09.4f,%c,%010.4f,%c,1,%s,%s,,*", 
	    pgga, time, lat, ltd, lon, lnd, sts, alt);
#endif
    fprintf(stream, "%s%02X\r\n", gga, string_checksum(gga));

    /* If fix includes extended fix quality information, construct
       GPGSA and GPGSV sentences. */
    if (lfp->fxtyp > 3) {
#ifdef EMULATE_RTK_TOOL
      int azmth = 0;
#endif
      int fnsat = 0;

      /* Construct GPGSA sentence */
      #if 1
      sprintf(gsa, "%s,A,%s,,,,,,,,,,,,,", pgsa, fxql);
      for (j = 0; j < 12; j++)
	if (fxp[n].sat[j * 2])
	  fnsat++;
      #else
      /* There does not appear to be any way of determining which
	 satellites are used for the fix from the stored data; the
	 approach here merely includes all satellites in view, which
	 is incorrect. */
      sprintf(gsa, "%s,A,%s,", pgsa, fxql);
      for (j = 0; j < 12; j++)
	if (fxp[n].sat[j * 2]) {
	  fnsat++;
	  sprintf(gsa + strlen(gsa), "%02d,", fxp[n].sat[2*j]);
	} else {
	  strcat(gsa, ",");
	}
      #endif
      sprintf(gsa + strlen(gsa), "%.1f,%.1f,%.1f*", pdop, hdop, vdop);
      #ifndef EMULATE_RTK_TOOL
      fprintf(stream, "%s%02X\r\n", gsa, string_checksum(gsa));
      #endif

      /* Construct GPGSV sentences */
      for (i = 1; i <= ((fnsat + 3) / 4); i++) {
	sprintf(gsv, "%s,%d,%d,%d", pgsv, ((fnsat + 3) / 4), i, fnsat);
	for (j = (i - 1) * 4; j < (i * 4); j++) {
	  if (!fxp[n].sat[j * 2])
	    continue;
#ifdef EMULATE_RTK_TOOL
	  sprintf(gsv + strlen(gsv), ",%02d,45,%03d,%d",  fxp[n].sat[j * 2], 
		  azmth, fxp[n].sat[j * 2 + 1]);
	  azmth += 30;
#else
	  sprintf(gsv + strlen(gsv), ",%02d,,,%d",  fxp[n].sat[j * 2], 
		  fxp[n].sat[j * 2 + 1]);
#endif
	}
	strcat(gsv, "*");
	fprintf(stream, "%s%02X\r\n", gsv, string_checksum(gsv));
      }
    }

    /* Construct GPRMC sentence */
#ifdef EMULATE_RTK_TOOL
    sprintf(rmc, "%s,%s,A,%09.4f,%c,%010.4f,%c,%s,%s,%2.2s%2.2s%2.2s,,,E*",
#else
    sprintf(rmc, "%s,%s,A,%09.4f,%c,%010.4f,%c,%s,%s,%2.2s%2.2s%2.2s,,,*",
#endif
	    prmc, time, lat, ltd, lon, lnd, vel, crs, lfp->date+6, 
	    lfp->date+4, lfp->date+2);
    fprintf(stream, "%s%02X\r\n", rmc, string_checksum(rmc));

    /* Construct RTDIST sentence (Royaltek extension, not standard NMEA) */
    if (lfp->fxtyp > 2) {
      if (lfp->fxtyp > 3)
	sprintf(rtd, "%s,A,%s,%.1f,%.1f,%.1f,%u*", prtd, fxql,
		pdop, hdop, vdop, fxp[n].dist);
      else
	sprintf(rtd, "%s,A,3,,,,%u*", prtd, fxp[n].dist);
      fprintf(stream, "%s%02X\r\n", rtd, string_checksum(rtd));
    }

#if 0
    fprintf(stream, "u:%02hx,h:%02hx,m:%02hx,s:%02hx,", (short)fxp[n].unkw0,
	    (short)fxp[n].hour, (short)fxp[n].min, (short)fxp[n].sec);
    fprintf(stream, "l:%e,l:%e,a:%e,v:%e,d:%08x,f:%02hx,n:%02hx,", fxp[n].lat,
	    fxp[n].lng, fxp[n].alt, fxp[n].vel, fxp[n].dist, 
	    (short)fxp[n].fstt, (short)fxp[n].nsat);
    fprintf(stream, "h:%04hx,p:%04hx,v:%04hx,", fxp[n].hdop, fxp[n].pdop, 
	    fxp[n].vdop);
    for (i = 0; i < 12; i++) {
      fprintf(stream, "s%02d:%02hx,%02hx,", i, (short)fxp[n].sat[2*i], 
	      (short)fxp[n].sat[2*i+1]);
    }
    fprintf(stream, "c:%e\n", fxp[n].crse);
#endif

  }
}


/*****************************************************************************
 Initialise geoid calculation structure.
 *****************************************************************************/
int geoid_calc_open(const char *fnam, geoid_height_t *gdhtp) {
  int fd;

  gdhtp->filep = NULL;
  gdhtp->gridp = NULL;
  if ((fd = open(fnam, O_RDONLY)) == -1)
    return -1;

  if (read(fd, gdhtp, 2*sizeof(uint16_t)+7*sizeof(float)) == -1)
    return -1;

#ifdef WORDS_BIGENDIAN
  /* If necessary, swap endianness of header values */
  bswap_16(*(__uint16_t*)(char*)&gdhtp->nlat);
  bswap_16(*(__uint16_t*)(char*)&gdhtp->nlng);
  bswap_32(*(__uint32_t*)(char*)&gdhtp->latmin);
  bswap_32(*(__uint32_t*)(char*)&gdhtp->latstp);
  bswap_32(*(__uint32_t*)(char*)&gdhtp->latmax);
  bswap_32(*(__uint32_t*)(char*)&gdhtp->lngmin);
  bswap_32(*(__uint32_t*)(char*)&gdhtp->lngstp);
  bswap_32(*(__uint32_t*)(char*)&gdhtp->lngmax);
  bswap_32(*(__uint32_t*)(char*)&gdhtp->qscale);
#endif

  gdhtp->filep = mmap(NULL, 2*sizeof(uint16_t)+7*sizeof(float) + 
		      gdhtp->nlat*gdhtp->nlng*sizeof(int16_t),
		      PROT_READ, MAP_PRIVATE, fd, 0);

  if (gdhtp->filep == MAP_FAILED) {
    gdhtp->filep = NULL;
    return -1;
  }

  if (close(fd))
    return -1;

  gdhtp->gridp = (int16_t *)((char *)gdhtp->filep + 
			     2*sizeof(uint16_t)+7*sizeof(float));
  return 0;
}


/*****************************************************************************
 Compute geoid correction with respect to WGS84 ellipsoid at specified
 location.
 *****************************************************************************/
float geoid_calc_correction(const geoid_height_t *gdhtp, float lat,float lng) {
  float slat, slng, x, y, x1, y1, g00, g01, g10, g11;
  unsigned short ilat0, ilng0, ilat1, ilng1;

  if (lat < gdhtp->latmin || lat > gdhtp->latmax)
    return 0.0/0.0;
  if (lng < gdhtp->lngmin || lng > gdhtp->lngmax)
    return 0.0/0.0;
  
  slat = (lat - gdhtp->latmin)/gdhtp->latstp;
  if ((ilat0 = floorf(slat + 0.5f)) == slat)
    ilat1 = ilat0;
  else {
    ilat0 = floorf(slat);
    ilat1 = ceilf(slat);
  }
  y = slat - ilat0;
  y1 = 1.0f - y;
  slng = (lng - gdhtp->lngmin)/gdhtp->lngstp;
  if ((ilng0 = floorf(slng + 0.5f)) == slng)
    ilng1 = ilng0;
  else {
    ilng0 = floorf(slng);
    ilng1 = ceilf(slng);
  }
  x = slng - ilng0;
  x1 = 1.0f - x;

#ifdef WORDS_BIGENDIAN
  {
    int16_t q00, q01, q10, q11;
    q00 = gdhtp->gridp[ilng0*gdhtp->nlat + ilat0];
    q01 = gdhtp->gridp[ilng1*gdhtp->nlat + ilat0];
    q10 = gdhtp->gridp[ilng0*gdhtp->nlat + ilat1];
    q11 = gdhtp->gridp[ilng1*gdhtp->nlat + ilat1];
    bswap_16(*(__int16_t*)(char*)&q00);
    bswap_16(*(__int16_t*)(char*)&q01);
    bswap_16(*(__int16_t*)(char*)&q10);
    bswap_16(*(__int16_t*)(char*)&q11);
    g00 = q00/gdhtp->qscale;
    g01 = q01/gdhtp->qscale;
    g10 = q10/gdhtp->qscale;
    g11 = q11/gdhtp->qscale;
  }
#else
  g00 = (gdhtp->gridp[ilng0*gdhtp->nlat + ilat0])/gdhtp->qscale;
  g01 = (gdhtp->gridp[ilng1*gdhtp->nlat + ilat0])/gdhtp->qscale;
  g10 = (gdhtp->gridp[ilng0*gdhtp->nlat + ilat1])/gdhtp->qscale;
  g11 = (gdhtp->gridp[ilng1*gdhtp->nlat + ilat1])/gdhtp->qscale;
#endif

  return g00*(x1*y1) + g01*(y1*x) + g10*(y*x1) + g11*(x*y);
}


/*****************************************************************************
 Destroy geoid calculation structure.
 *****************************************************************************/
int geoid_calc_close(geoid_height_t *gdhtp) {
  if (gdhtp->filep != NULL) {
    if (munmap(gdhtp->filep, 2*sizeof(uint16_t)+7*sizeof(float) + 
	       gdhtp->nlat*gdhtp->nlng*sizeof(int16_t)) == -1)
      return -1;
    gdhtp->filep = NULL;
  }
  gdhtp->gridp = NULL;
  return 0;
}
